/* server.c: recipient of forwarded calls */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "netproto.h"
#include "ioctls.h"

#define MAX_BUFFER_LEN 10485760    /* more than 10MB of buffer is insane */

int fd, sv_sock, sock;
int mbufsize;
char *buf;

const char *fwd_allowed[] = FWD_FILES;

int is_eligible(const char *file)
{
  int i;

  for (i=0; fwd_allowed[i] != NULL; i++)
    if (!strcmp(file, fwd_allowed[i])) return 1;
  return 0;
}

void process_connection()
{
  int req_code, flag, arg, len, result;
  int ioctl_props;
  char *fname;

  fd = -1;
  mbufsize = 2048;
  buf = (char *) malloc(mbufsize);
  if (buf == NULL) return;
  arg = 1;
  setsockopt(sock, SOL_TCP, TCP_NODELAY, &arg, sizeof(arg));
  
  while (1) {
    if (!recv_int(sock, &req_code)) return;
    if ((fd < 0) && (req_code != CALL_OPEN)) return; /* protocol error */
    
    switch(req_code) {
    
      case CALL_OPEN:
        recv_string(sock, &fname);
        recv_int(sock, &flag);
        if (flag & O_CREAT) recv_int(sock, &arg);
        else arg = 0;
        /* is file forwardable ? */
        if ((fd >=0) || !is_eligible(fname)) 
          { send_int(sock, -1); flush_send(sock); return; }
        fd = open(fname, flag, arg);
        send_int(sock, fd);
        flush_send(sock);
        if (fd == -1) return;
        break;

      case CALL_CLOSE:
        result = close(fd);
        send_int(sock, result);
        flush_send(sock);
        return;
        
      case CALL_WRITE:
        recv_int(sock, &len);
        if (len > MAX_BUFFER_LEN) return; /* we're being DoS'ed */
        if (len > mbufsize) { 
          mbufsize = len;
          buf = (char *) realloc(buf, len);
          if (buf == NULL) return;
        }
        recv_data(sock, len, buf);
        result = write(fd, buf, len);
        send_int(sock, result);
        flush_send(sock);
        break;
        
      case CALL_READ:
        recv_int(sock, &len);
        if (len > MAX_BUFFER_LEN) return; /* we're being DoS'ed */
        result = read(fd, buf, len);
        send_int(sock, result);
        if (result > 0) send_data(sock, len, buf);
        flush_send(sock);
        break;
        
      case CALL_LSEEK:
        recv_int(sock, &len);
        recv_int(sock, &flag);
        result = lseek(fd, len, arg);
        send_int(sock, result);
        flush_send(sock);
        break;
        
      case CALL_IOCTL:
        recv_int(sock, &arg);
        if (!prepare_ioctl_info(arg, &ioctl_props, &len)) return;
        if (len > mbufsize) { 
          mbufsize = len;
          buf = (char *) realloc(buf, len);
          if (buf == NULL) return;
        }
        if (ioctl_props & PROPS_WRITE) recv_data(sock, len, buf);
        result = ioctl(fd, arg, buf);
        if (ioctl_props & PROPS_READ)  send_data(sock, len, buf);
        send_int(sock, result);
        flush_send(sock);
        break;
        
      default: /* unknown */
        return;
    }
  }
}

int main (int argc, char *argv[])
{
  struct sockaddr_in saddr;
  int i;
  
  /* init listen socket */
  saddr.sin_family = AF_INET;
  saddr.sin_addr.s_addr = INADDR_ANY;
  saddr.sin_port = htons(argc>1 ? atoi(argv[1]) : DEFAULT_PORT);
  sv_sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sv_sock < 0) exit(1);
  if (bind(sv_sock, &saddr, sizeof saddr) < 0)
  {
    printf("iaudioserv: cannot bind to port %d. Try running \"%s 0\"\n",
      ntohs(saddr.sin_port), argv[0]);
    exit(1);
  }
  i = sizeof saddr;
  getsockname(sv_sock, &saddr, &i);
  printf("iaudioserv: listening on port %d.\n", ntohs(saddr.sin_port));
  listen(sv_sock, 20);

  /* main loop */  
  while (1) {
    sock = accept(sv_sock, NULL, NULL);
    if (fork()) {
      /* parent completes the job */
      close(sv_sock);
      process_connection();
      close(sock);
      exit(1);
    }
    else close(sock);
  }
}